/***************************************************************************
* Copyright (c) 2017, Sylvain Corlay and Johan Mabille                     *
*                                                                          *
* Distributed under the terms of the BSD 3-Clause License.                 *
*                                                                          *
* The full license is in the file LICENSE, distributed with this software. *
****************************************************************************/

#ifndef XWIDGETS_CONFIG_CLING_HPP
#define XWIDGETS_CONFIG_CLING_HPP

#pragma cling add_library_path("C:/bld/xwidgets_1545634336024/_h_env/Library/lib")
#pragma cling load("libxwidgets")

#endif
